"""The 'worktoy.keenum' package provides the KeeNum class for enumeration!"""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._num import Num, auto
from ._space_num import SpaceNum
from ._meta_num import MetaNum
from ._keenum import KeeNum
